/**************************************************************************************************************************
// This script was authored by Lorne Brooks - Email: lbtools2k@gmail.com
//
//
// REVISION HISTORY:
// Version 1.3 27/11/18;
//						- Added a mode selection to apply offsets to controllers for position, rotation or both.
//						- Added an option to invert the difference values when applying them.
//						- Improved user interface feedback, clarity and usage.
//						- General code cleanup and refactoring.
//						- Added a toolbar icon.
// Version 1.2 27/10/14;
//						- Email change for help contact.
//						- Recoded script as a struct to minimize global variables.
// Version 1.1 17/06/08;
//						- Changed the way controller variables are stored - script should actually work now!
// Version 1.0 16/05/08;
//						- First public release
//
//
// Description:
// This script functions similar to Biped's own Apply Increment tool. Basically, it calculates the difference between initial
// position or rotation values at a given time and applies the difference in values to any selected keys on selected objects.
//
// Usage:
// 1. Select a single object and click Get Initial Values to record it's position and orientation values at that time.
// 2. Move or rotate the object (animated or not) and click Get New Values button to record the new values.
// 3. Select some objects and keys (in trackview or trackbar) then click >>> Apply NOW <<<.
//		The calculated difference in position and rotation values between the Initial and New are applied relative to the
//		selected key values as an offset.
//
// Limitations:
// - Currently, only works on Position/Rotation/Scale controllers and IK chain controllers for position and rotation tracks.
// - Doesn't work on Biped. It has its own version of Apply Increment (the initial inspiration for this tool).
// - Doesn't work on CAT rigs.
//
// Known bugs:
// A crash may occur if attempting to use with IK solvers. You can still use it, just save your work after applying the script.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/

global g_LBTools_AI

struct s_LBTools_applyIncrement
(

/*==================================================================================
// Structure variables
====================================================================================*/

applyIncrement_Rollout,
aboutScript_rollout,
dPos,

/*==================================================================================*/
	
fn f_applyIncrement =
(
	rollout applyIncrement_Rollout "Apply Increment"
	(
		local iniPos = undefined
		local iniRot = undefined
		local newPos = undefined
		local newRot = undefined
		local changePos = undefined
		local changeRot = undefined

		fn f_applyToObjects ctrlsState isInverted =
		(
			local a_selObjs = selection as array
			if a_selObjs.count > 0 then
			(
				local keysWereSelected = false
				
				for obj in a_selObjs do
				(
					--// Apply the offsets to the selected keys
					with animate on
					(	
						-- // ORIENTATION CHANGE MUST COME FIRST
						if ctrlsState == 2 OR ctrlsState == 3 do
						(
							-- // Apply to selected rotation keys
							for i = 1 to obj.rotation.controller.keys.count do
							(
								local result = isKeySelected obj.rotation.controller i
								if result then
								(
									sliderTime = obj.rotation.controller.keys[i].time
									if classof obj.controller == PRS do
									(
										if isInverted then
											obj.controller.rotation -= changeRot
										else
											obj.controller.rotation += changeRot
									)
									if classof obj.controller == IKChainControl do
									(
										if isInverted then
											obj.transform.controller.ik_goal.controller.rotation -= changeRot
										else
											obj.transform.controller.ik_goal.controller.rotation += changeRot
									)
									keysWereSelected = true									
								)
							)
						)
						if ctrlsState == 1 OR ctrlsState == 3 do
						(
							-- // Apply to selected position keys
							for i = 1 to obj.position.controller.keys.count do
							(
								local result = isKeySelected obj.position.controller i
								if result do
								(
									sliderTime = obj.position.controller.keys[i].time
									if classof obj.controller == PRS do
									(
										if isInverted then
											obj.controller.position -= changePos
										else
											obj.controller.position += changePos
									)
									if classof obj.controller == IKChainControl do
									(
										if isInverted then
											obj.transform.controller.ik_goal.controller.position -= changePos
										else
											obj.transform.controller.ik_goal.controller.position += changePos
									)
									keysWereSelected = true
								)
							)
						)
					)
				)
				
				if NOT keysWereSelected do
				(
					messagebox "Error: No keys selected for any of the selected objects.\n" title:"Apply Increment"
				)
			)
			else
			(
				messagebox "Error: No objects selected." title:"Apply Increment"
			)
		)
		
		group "Stored Values"
		(
			button btn_getInitial "Get INITIAL Values" width:150 height:25 across:2 toolTip:"Store the INITIAL position and rotation values at the current time"
			button btn_getNew "Get NEW Values" width:150 height:25 toolTip:"Store the NEW position and rotation values at the current time"
			edittext edtxt_iniPos "Ini Pos:    " align:#left width:320 readonly:true
			edittext edtxt_iniRot "Ini Rot:    " align:#left width:320 readonly:true
			edittext edtxt_newPos "New Pos: " align:#left width:320 readonly:true
			edittext edtxt_newRot "New Rot: " align:#left width:320 readonly:true
		)
		group "Apply To Selected Objects"
		(
			label lbl_note "Note: Only supports PRS and IKChainControl controllers."
			label lbl_gap1 ""
			label lbl_txt "Affected Controllers:" align:#left across:2
			radiobuttons rdobtns_ctrls "" labels:#("Position","Rotation","Both") offsets:#([-10,0],[0,0],[10,0]) columns:3 default:3 align:#left offset:[-40,0]
			label lbl_gap2 ""
			checkbox chkbx_inverted "Invert difference values" align:#left checked:false
			button btn_applyIncrement ">>>   Apply NOW   <<<" width:300 height:30 toolTip: "Apply the calculated difference between the INITIAL and NEW values to the selected keys for all selected objects"
		)
		button btn_about "About" align:#center
		
		on btn_about pressed do
		(
			try(destroyDialog g_LBTools_AI.aboutScript_rollout)catch()
			createDialog g_LBTools_AI.aboutScript_rollout
		)
		
		on applyIncrement_Rollout close do
		(
			try(destroyDialog g_LBTools_AI.aboutScript_rollout)catch()
		)
		
		on applyIncrement_Rollout moved val do
		(
			g_LBTools_AI.dPos = val
		)
		
		on btn_getInitial pressed do
		(
			if selection.count == 1 then
			(
				local obj = selection[1]
				if classof obj.controller == PRS do
				(
					iniPos = obj.controller.position
					iniRot = obj.controller.rotation
				)
				if classof obj.controller == IKChainControl do
				(
					iniPos = obj.controller.ik_goal.controller.position
					iniRot = obj.controller.ik_goal.controller.rotation
				)
				edtxt_iniPos.text = iniPos as string
				edtxt_iniRot.text = iniRot as string
			)
			else
			(
				messagebox "Error: Select a single object." title:"Apply Increment"
			)
		)

		on btn_getNew pressed do
		(
			if selection.count == 1 then
			(
				local obj = selection[1]
				if classof obj.controller == PRS do
				(
					newPos = obj.controller.position
					newRot = obj.controller.rotation
				)
				if classof obj.controller == IKChainControl do
				(
					newPos = obj.controller.ik_goal.controller.position
					newRot = obj.controller.ik_goal.controller.rotation
				)
				edtxt_newPos.text = newPos as string
				edtxt_newRot.text = newRot as string
			)
			else
			(
				messagebox "Error: Select a single object." title:"Apply Increment"
			)
		)

		on btn_applyIncrement pressed do
		(
			if iniPos != undefined AND iniRot != undefined AND newPos != undefined AND newRot != undefined then
			(
				undo "ApplyIncrement" on
				(
					--// Calculate the differences
					changePos = newPos - iniPos
					changeRot = newRot - iniRot
					
					f_applyToObjects rdobtns_ctrls.state chkbx_inverted.state
				)
			)
			else
			(
				messagebox "Error: Both of the INITIAL and NEW values must be set." title:"Apply Increment"
			)
		)
	) --// end rollout

	rollout aboutScript_rollout "About"
	(
		label lbl_author "Author - Lorne Brooks"
		hyperLink myEmail "lbtools2k@gmail.com" color:(color 0 0 255) hovercolor:(color 0 0 190) visitedcolor:(color 190 0 0) align:#center address:"mailto:lbtools2k@gmail.com"
		label lbl_version "Version 1.3 - 27 Nov 2018"
	) --// end rollout
	
	--// Load UI
	try(destroyDialog applyIncrement_Rollout)catch()
	if dPos == undefined do
	(
		dPos = mouse.screenpos + [-75,-15]
	)
	createDialog applyIncrement_Rollout width:350 pos:dPos
) --// end function
	
) --// end struct

g_LBTools_AI = s_LBTools_applyIncrement()


/*********************************************************************************************************************************************************************/
--// MACRO CALLS FOR UI
/*********************************************************************************************************************************************************************/

macroscript applyIncrement
	category:"LB Tools"
	buttonText: "{ APP-INC }"
	icon:#("LBTools_ApplyIncrement",1)
	toolTip:"Apply Increment 1.3"
(
	g_LBTools_AI.f_applyIncrement()
)

/*********************************************************************************************************************************************************************/